/*

                                                                                       
 File name:   KERNEL.H                                                                 
 Description: System Control Blocks definitions for OS/2 2.0                           
 Author:      Luca Fiori                                                               
 Date:        September 1992                                                           
 Notes:       1 - SAS segment is pointed by the SAS_SELECTOR:SAS_OFFSET pair           
              2 - ppPTDAChain don't points to the PTDA's chain directly, but it points 
                  to a linear location that contains the linear address of that chain  
              3 - For see if a chain is ended use the beginners and the terminators of 
                  the chain:                                                           
                      DEV_LASTDDOFF                                                    
                      DEV_LASTDDSEG                                                    
                      PTDA_FIRSTPTDA                                                   
                      PTDA_LASTPTDA                                                    
                      TCB_LASTTCB                                                      
                      MTE_LASTMTE                                                      
              4 - When a system control block is reached, verify its validity by       
                  controlling its signature:                                           
                      SAS_SIGNATURE                                                    
                      PTDA_SIGNATURE                                                   
                      MTE16_SIGNATURE                                                  
                      MTE32_SIGNATURE                                                  
              5 - The main links diagram (verified by me) is the following:            
                                                                                       
                                                                    ͻ    
                  ͻ                                ʻ<Ŀ
                   SAS                                            ʻ 
                     usDDChainOff> DDHEADER     
                     ppPTDAChainĿ                            ȹ   usOffNextĴ
                     pKMTEChainĿ                              ȹ   usSelNext
                  ͼ    ͻ         ͼ  
                                        ʻ<Ŀ                      
                                        ʻ                       
                                         PTDA                ͻ    
                                      >   pFirstChildĴ   ʻ<Ŀ
                                            pNextSibling   ʻ 
                                         ȹ   pFirstTCB>ȹ TCB          
                                          ȹ   hModuleĿ    ȹ   pNextTCB
                                           ͼ      ͼ  
                                                                                     
                                          ͻ         ͻ    
                                          ʻ<Ŀ     ʻ<Ŀ
                                          ʻ      ʻ 
                                     > MTE                MTE          
                                           ȹ   pNext     ȹ   pNext
                                            ȹ                 >hModule     
                                             ͼ          ͼ  
                                                                                       

*/




/*
ͻ
 Module Table Entry resident portion definition                                        
ͼ
*/
#define MTE16_SIGNATURE "NE"            /* Signature for 16-bit module                 */
#define MTE32_SIGNATURE "LX"            /* Signature for Linear eXecutable module      */
#define MTE_LASTMTE     0x00000000      /* Last MTE pointer in chain is 00000000       */

typedef struct _MTE
        {
            CHAR        chSignature[2]; /* 000: Signature: "NE" or "LX"                */
            USHORT      usUseCount;     /* 002: Use count                              */
            ULONG       ulFlags;        /* 004: Flags                                  */
            ULONG       ulMNTEntr;      /* 008: # entries in Imp. Module Table         */
            struct
            {
                BYTE    bLength;        /* 000: Module name length                     */
                CHAR    chName;         /* 001: Module name                            */
            } *pName;
            PVOID       pModPtrsTbl;    /* 010: Pointer to module pointers table       */
            USHORT      hModule;        /* 014: Module handle (HOB)                    */
            USHORT      usFSNumber;     /* 016: File system # for open files           */
            struct _MTE *pNextMTE;      /* 018: Pointer to next MTE                    */
            PVOID       pSwapMTE;       /* 01C: Pointer to swap portion of MTE         */
        } MTE, *PMTE;




/*
ͻ
 System Anchor Segment definition                                                      
ͼ
*/
#define SAS_SIGNATURE "SAS "            /* Signature of SAS (at the begin)             */
#define SAS_SELECTOR  0x0070            /* Segment selector of SAS                     */
#define SAS_OFFSET    0x0000            /* Segment offset of begin of SAS              */

typedef struct _SAS
        {
            CHAR        chSignature[4]; /* 000: Signature: "SAS "                      */
            CHAR        chBoh0[28];     /* 004: ??? <========                          */

            /***************************************************************************/
            /* Device Driver area                                                      */
            /***************************************************************************/
            USHORT      usDDChainOff;   /* 020: Offset of first DD header here in SAS  */
            BYTE        bBoh1[10];      /* 022: ??? <========                          */

            /***************************************************************************/
            /* Memory Management area                                                  */
            /***************************************************************************/
            BYTE        bBoh2[4];       /* 02C: ??? <========                          */
            PVOID       ppObjRecord;    /* 030: Pointer to address of object record    */
            PVOID       ppCtxtRecord;   /* 034: Pointer to address of context record   */
            PMTE        pKMTEChain;     /* 038: Pointer to kernel MTE records chain    */
            PMTE        pLMTEList;      /* 03C: Pointer to linked MTE list             */
            PVOID       ppPgFrmTbl;     /* 040: Pointer to address of page frame table */
            PVOID       ppPgRngTbl;     /* 044: Pointer to address of page range table */
            struct
            {
                PVOID   pSwapTbl;       /* 000: Pointer to swap frame table            */
                ULONG   ulSwapTblLen;   /* 004: Swap table length                      */
            } *pSwap;
            PVOID       ppIdleHead;     /* 04C: Pointer to address of idle head        */
            PVOID       ppFreeHead;     /* 050: Pointer to address of free head        */
            BYTE        bBoh3[4];       /* 054: ??? <========                          */

            /***************************************************************************/
            /* Task area                                                               */
            /***************************************************************************/
            USHORT      usBoh4;         /* 058: ??? <========                          */
            PVOID       ppPTDAChain;    /* 05A: Pointer to address of PTDA chain       */
            PVOID       ppThreadArray;  /* 05E: Pointer to address of thread array     */
            PVOID       pCurrThread;    /* 062: Pointer to current thread              */
            PVOID       pThreadCount;   /* 066: Pointer to thread count                */
        } SAS, *PSAS;




/*
ͻ
 Device Driver header                                                                  
ͼ
*/
#define DEV_LASTDDOFF 0xFFFF            /* Last DD head offset in chain is FFFF        */
#define DEV_LASTDDSEL 0xFFFF            /* Last DD head selector in chain is FFFF      */

#define DEV_CHAR_DEV  0x8000            /* C15 - Character devices:                    */
#define DEV_BLOCK_DEV 0x0000            /* B15 - Block devices:                        */
#define DEV_IOCTL     0x4000            /*  14 -   understands IOCTL strings           */
#define DEV_OUTTILBSY 0x2000            /* C13 -   supports output-until-busy          */
#define DEV_FAT_ID    0x2000            /* B13 -   examines the FAT ID byte            */
#define DEV_OPENCLOSE 0x0800            /*  11 -   understands Open/Close              */
#define DEV_NULL1     0x0000            /*  10 -   must be 0                           */
#define DEV_NULL2     0x0000            /*  09 -   must be 0                           */
#define DEV_PRG_IOCTL 0x0100            /*  08 -   und.Pause/Restart/Gen.IOCTL         */
#define DEV_INT29H    0x0010            /* C04 -   is recipient of INT 29h             */
#define DEV_CLOCK     0x0008            /* C03 -   is clock device                     */
#define DEV_NULL      0x0004            /* C02 -   is null device                      */
#define DEV_CONS_OUT  0x0002            /* C01 -   is console output                   */
#define DEV_CONS_IN   0x0001            /* C00 -   is console input                    */

typedef struct _DDHEADER
        {
            USHORT usOffNext;           /* 000: Offset to next header                  */
            USHORT usSelNext;           /* 002: Selector to next DD header             */
            USHORT usAttribute;         /* 004: DD attribute                           */
            USHORT usStrategy;          /* 006: Offset of strategy  routine            */
            USHORT usInterrupt;         /* 008: Offset of interrupt routine            */
            CHAR   chName[8];           /* 00A: DD name (8 ASCII chars)                */
            USHORT usProtCS;            /* 012: Reserved 1 - CS protected mode         */
            USHORT usProtDS;            /* 014: Reserved 2 - DS protected mode         */
            USHORT usRealCS;            /* 016: Reserved 3 - CS real      mode         */
            USHORT usRealDS;            /* 018: Reserved 4 - DS real      mode         */
        } DDHEADER, *PDDHEADER;




/*
ͻ
 Thread Control Block definition                                                       
ͼ
*/
#define TCB_LASTTCB 0x00000000          /* Last TCB pointer in chain is 00000000       */

#define TS_ZOMBIE   0                   /* Thread's states defines for 00E0            */
#define TS_READY    1                   /* and 00E1 fields:                            */
#define TS_BLOCKED  2                   /*     - bSchedQueLoc                          */
#define TS_CRITSEC  4                   /*     - bSchedStat                            */
#define TS_RUNNING  5

/***************************************************************************************/
/* Thread's priority classes, as defined in the BSEDOS.H file (useful for the 00E4     */
/* field (bPrtyClass)):                                                                */
/*                                                                                     */
/* #define PRTYC_IDLETIME         1     Idle-Time class                                */
/* #define PRTYC_REGULAR          2     Regular class                                  */
/* #define PRTYC_TIMECRITICAL     3     Time-Critical class                            */
/* #define PRTYC_FOREGROUNDSERVER 4     Fixed-High or Foreground Server class          */
/***************************************************************************************/

typedef struct _TCB
        {
            /***************************************************************************/
            /* TCB link informations                                                   */
            /***************************************************************************/
            TID         tidTID;         /* 000: Thread IDentificator                   */
            USHORT      usThreadNum;    /* 002: Thread slot number                     */
            ULONG       ulForcedAct;    /* 004: Forced action bit vector               */
            PVOID       pPTDA;          /* 008: Pointer to PTDA (PPTDA cast)           */
            PVOID       pTSD;           /* 00C: Thread Swappable Data                  */
            PVOID       pTIB;           /* 010: Thread Information Block               */
            struct _TCB *pNextTCB;      /* 014: Next TCB                               */

            /***************************************************************************/
            /* Dos32CreateThread stack informations                                    */
            /***************************************************************************/
            ULONG       ulVirStackSz;   /* 0018: Virtual stack object size             */
            ULONG       ulComStackSz;   /* 001C: Committed stack object size           */
            ULONG       ulVirStackBs;   /* 0020: Virtual stack object base             */
            PVOID       pLLDArLkHd;     /* 0024: Libi load data arena link head        */
            PVOID       pLLDArLkCr;     /* 0028: Libi load data arena link current     */
            PVOID       pLLDArLkFr;     /* 002C: Libi load data arena link free        */
            BYTE        bBoh0[8];       /* 0030: ??? <========                         */
            ULONG       ulStackFrTyp;   /* 0038: Stack frame type                      */
            ULONG       ulStackFrBP;    /* 003C: Stack frame base pointer              */
            ULONG       ulBoh1;         /* 0040: ??? = at first of the next            */
            ULONG       ulLCtHkHd[2];   /* 0044: Local context hook head               */

            /***************************************************************************/
            /* Information needed for context switching                                */
            /***************************************************************************/
            ULONG       ulKStPg0TCB;    /* 004C: KStack page 0 of TCB                  */
            ULONG       ulKStPg1TCB;    /* 0050: KStack page 1 of TCB                  */
            ULONG       ulKStPgTSD;     /* 0054: KStack page of TSD                    */
            ULONG       ulKStPg0PTDA;   /* 0058: KStack page 0 of PTDA                 */
            ULONG       ulKStPg1PTDA;   /* 005C: KStack page 1 of PTDA                 */
            ULONG       ulKStPg2PTDA;   /* 0060: KStack page 2 of PTDA                 */
            ULONG       ulSSOffTCB;     /* 0064: SS-relative offset to current TCB     */
            ULONG       ulSSOffTSD;     /* 0068: SS-relative offset to current TSD     */
            ULONG       ulStToFltTCB;   /* 006C: Stack-to-flat TCB conversion value    */
            ULONG       ulStToFltTSD;   /* 0070: Stack-to-flat TSD conversion value    */

            /***************************************************************************/
            /* Variables used in the file system                                       */
            /***************************************************************************/
            ULONG       ulUsrIOTrfAd;   /* 0074: User's I/O transfer address           */
            ULONG       ulFrSectAcc;    /* 0078: First sector accessed in file         */
            PVOID       pSFT;           /* 007C: SFT we're working with                */
            ULONG       ulValidSects;   /* 0080: Valid (previous written) sectors      */
            PVOID       pRedirTCB;      /* 0084: Redirector TCB(used by LANMAN)        */
            USHORT      usPIDFSC;       /* 0088: Process ID for file share checks      */
            USHORT      usUIDFSC;       /* 008A: User ID for file share checks         */
            BYTE        bRedir;         /* 008C: Redirection: 0=no, 1=yes              */

            BYTE        bBoh2[67];      /* 008D: ??? <=== I will find!                 */

            ULONG       ulDebugKsems;   /* 00D0: Debugger displacement addr for ksems  */

            /***************************************************************************/
            /* Per-thread NPX information                                              */
            /***************************************************************************/
            ULONG       ulNPXBuffer;    /* 00D4: Buffer                                */

            /***************************************************************************/
            /* TKWaitThread fields                                                     */
            /***************************************************************************/
            PVOID       pNextWaiting;   /* 00D8: Next waiting TCB                      */
            ULONG       ulWaitsMeDie;   /* 00DC: Threads waiting for me to die         */

            /***************************************************************************/
            /* Scheduler fields                                                        */
            /***************************************************************************/
            BYTE        bSchedQueLoc;   /* 00E0: Scheduler queue location (actual)     */
            BYTE        bSchedStat;     /* 00E1: Current scheduler state (desired)     */
            BYTE        bTKSlpWkpFlg;   /* 00E2: TKSleep/TKWakeup flags                */
            BYTE        bWndBoostCnt;   /* 00E3: Window boost count                    */
            BYTE        bPrtyClass;     /* 00E4: Priority class (user)                 */
            BYTE        bPrtyLevel;     /* 00E5: Priority level (user)                 */
            BYTE        bPrtyClassMd;   /* 00E6: Priority class modifier bits          */
            BYTE        bMscSchedFlg;   /* 00E7: Misc. scheduler flags                 */
            USHORT      usCalcPrty;     /* 00E8: Calculated priority                   */
            USHORT      usMinSchPrty;   /* 00EA: Minimum scheduling priority           */
            ULONG       ulKBoLkNsCnt;   /* 00EC: Kernel boost lock nesting count       */
            PVOID       pNextPrtyQue;   /* 00F0: Next priority queue in chain          */
            PVOID       pPrevPrtyQue;   /* 00F4: Previous priority queue in chain      */
            PVOID       pHighPrtyThr;   /* 00F8: Higher priority thread                */
            PVOID       pLowPrtyThr;    /* 00FC: Lower priority thread                 */
            PVOID       pNextThread;    /* 0100: Next same-priority thread             */
            PVOID       pPrevThread;    /* 0104: Previous same-priority thread         */
            PVOID       pTKQueryWkup;   /* 0108: TKQueryWakeup TCB list                */
            ULONG       ulSleepID;      /* 010C: SleepID of this TCB                   */

            BYTE        bBoh3[19];      /* 0110: ??? <========                         */

                                        /* Timeout/Starvation Timeout element          */
                                        /* Used by loader (TCBCheckedSig)(BYTE)        */

            /***************************************************************************/
            /* Per Thread Swapper information                                          */
            /***************************************************************************/
                                        /* Status of swapping (BYTE)                   */

            /***************************************************************************/
            /* Info about nexting level in FSH_DoVol0                                  */
            /***************************************************************************/
                                        /* Nexting level of FSH_DoVol0 (BYTE)          */

            /***************************************************************************/
            /* Device I/O fields                                                       */
            /***************************************************************************/
            BYTE        bReqPktFlag;    /* 0123: Flag indicating if request in use     */
            PVOID       pIOReqPKt;      /* 0124: I/O request packet for thread         */
        } TCB, *PTCB;




/*
ͻ
 Per Task Data Area definition                                                         
ͼ
*/
#define PTDA_SIGNATURE "TD"             /* Signature of PTDA (at the end)              */
#define PTDA_FIRSTPTDA 0x00000000       /* First PTDA ptr in chain is 00000000         */
#define PTDA_LASTPTDA  0x00000000       /* Last PTDA ptr in chain is 00000000          */

typedef struct _PTDA
        {
            PVOID       pParent;        /* 000: Parent PTDA                            */
            struct _PTDA *pPTDA;        /* 004: This PTDA                              */
            struct _PTDA *pFirstChild;  /* 008: First child PTDA                       */
            PVOID       pExecChild;     /* 00C:                                        */
            struct _PTDA *pNextSibling; /* 010: Next sibling PTDA                      */
            struct _PTDA *pPrevSibling; /* 014: Previous sibling PTDA                  */
            PSZ         szExeFile;      /* 018: Fully qualified .EXE file name         */
            PTCB        pLastTCB;       /* 01C: Last TCB in the TCB chain              */
            PTCB        pFirstTCB;      /* 020: First TCB in the TCB chain             */
            USHORT      usThreads;      /* 024: Number of threads                      */
            USHORT      usTIBCount;     /* 026: Number of allocated TIBs               */
            PVOID       pTIB0;          /* 028: First group of TIB's                   */
            PVOID       pTIB1;          /* 02C: Second group of TIB's                  */
            PVOID       pTIB2;          /* 030: Third group of TIB's                   */
            PVOID       pTIB3;          /* 034: Fourth group of TIB's                  */
            PVOID       pDebugData;     /* 038: Debug data???                          */
            ULONG       ulDebugFlag;    /* 03C: Debug flag???                          */

            /***************************************************************************/
            /* Virtual Memory Manager information                                      */
            /***************************************************************************/
            PVOID       pNextArena;     /* 040: Next arena header                      */
            PVOID       pPrevArena;     /* 044: Previous arena header                  */
            PVOID       pArenaSentin;   /* 048: Sentinel arena record                  */
            PVOID       pFirstFree;     /* 04C: Hint of first free                     */
            PVOID       pBitmapDir;     /* 050: Bitmap dir???                          */
            PVOID       pHashTable;     /* 054: Hash table???                          */
            PVOID       pPerTypeInfo;   /* 058: Per-type informations???               */
            ULONG       ulVMFlags;      /* 05C: VM flags???                            */
            ULONG       ulMinAddress;   /* 060: Min. address used by process           */
            ULONG       ulMaxAddress;   /* 064: Max. address used by process           */
            ULONG       ulArenaCount;   /* 068: Arena count???                         */
            ULONG       ulNeedBitmap;   /* 06C: Need bitmap???                         */
            ULONG       ulMaxBitmap;    /* 070: Max. bitmap???                         */
            ULONG       ulMaxBMEntry;   /* 074: Max. bitmap entry???                   */
            ULONG       ulMaxHashNdx;   /* 078: Max. hash index???                     */
            USHORT      usArenaHob;     /* 07C: Hob of pseu-obj in arena head.         */
            BYTE        bBoh0[2];       /* 07E: ??? <========                          */

            /***************************************************************************/
            /* Physical Memory Manager information                                     */
            /***************************************************************************/
            PVOID       pPageTables;    /* 080: Start of the page tables               */
            PVOID       pPageDir;       /* 084: Process page directory                 */
            PVOID       pAllocPage;     /* 088: Allocated page counts table            */
            PVOID       pPresPages;     /* 08C: Present page counts table              */
            PVOID       pLockResPage;   /* 090: Locked/resident page counts table      */
            ULONG       ulBaseVirtPg;   /* 094: Base virtual page number???            */
            ULONG       ulVDMAliasRg;   /* 098: Base VDM alias region???               */
            USHORT      usMaxPDES;      /* 09C: Max. potential pdes in arena           */
            USHORT      usLowPDES;      /* 09E: Count of low  in-use pdes              */
            USHORT      usHighPDES;     /* 0A0: Count of high in-use pdes              */
            USHORT      usPgTblCntxt;   /* 0A2: Page table context                     */
            USHORT      usPgMgrFlags;   /* 0A4: Per-process page manager flags         */

            USHORT      usEnvirHOB;     /* 0A6: Environment hob                        */
            BYTE        bBoh1[840];     /* 0A8: ??? <========                          */
            PVOID       pExitList;      /* 3F0: Exit List                              */
            BYTE        bBoh2[538];     /* 3F4: ??? <========                          */
            ULONG       ulPID;          /* 60E: Process IDentity (32 bit)              */
            BYTE        bBoh3[186];     /* 612: ??? <========                          */
            USHORT      hPTDA;          /* 6CC: Handle of PTDA                         */
            USHORT      hModule;        /* 6CE: Handle of module                       */
            USHORT      hLDT;           /* 6D0: Handle of LDT                          */
            BYTE        bBoh4[2];       /* 6D2: ??? <========                          */
            PVOID       pLDT;           /* 6D4: Local Descriptor Table                 */
            BYTE        bBoh5[8];       /* 6D8: ??? <========                          */
            USHORT      usOffFileTbl;   /* 6E0: Offset to open file table              */
            USHORT      usSelFileTbl;   /* 6E2: Selector to open file table            */
            USHORT      usOffBoh0;      /* 6E4: Offset to ??? <========                */
            USHORT      usSelBoh0;      /* 6E6: Selector to ??? <========              */
            BYTE        bBoh6[6];       /* 6E8: ??? <========                          */
            USHORT      usPID;          /* 6EE: Process IDentity (16 bit)              */
            USHORT      usParentPID;    /* 6F0: Parent's PID (16 bit)                  */
            BYTE        bBoh7[38];      /* 6F2: ??? <========                          */
            USHORT      usSelector;     /* 718: Selector                               */
            BYTE        bBoh8[8];       /* 71A: ??? <========                          */
            CHAR        chSignature[2]; /* 722: Signature: "TD"                        */
        } PTDA, *PPTDA;
